use crate::commands::WholeStreamCommand;
use crate::data::{config, Value};
use crate::errors::ShellError;
use crate::parser::hir::SyntaxType;
use crate::parser::registry::{self};
use crate::prelude::*;
use std::iter::FromIterator;
use std::path::PathBuf;

pub struct Config;

#[derive(Deserialize)]
pub struct ConfigArgs {
    load: Option<Tagged<PathBuf>>,
    set: Option<(Tagged<String>, Tagged<Value>)>,
    get: Option<Tagged<String>>,
    clear: Tagged<bool>,
    remove: Option<Tagged<String>>,
    path: Tagged<bool>,
}

impl WholeStreamCommand for Config {
    fn name(&self) -> &str {
        "config"
    }

    fn signature(&self) -> Signature {
        Signature::build("config")
            .named("load", SyntaxType::Path)
            .named("set", SyntaxType::Any)
            .named("get", SyntaxType::Any)
            .named("remove", SyntaxType::Any)
            .switch("clear")
            .switch("path")
    }

    fn usage(&self) -> &str {
        "Configuration management."
    }

    fn run(
        &self,
        args: CommandArgs,
        registry: &registry::CommandRegistry,
    ) -> Result<OutputStream, ShellError> {
        args.process(registry, config)?.run()
    }
}

pub fn config(
    ConfigArgs {
        load,
        set,
        get,
        clear,
        remove,
        path,
    }: ConfigArgs,
    RunnableContext { name, .. }: RunnableContext,
) -> Result<OutputStream, ShellError> {
    let name_span = name;

    let configuration = if let Some(supplied) = load {
        Some(supplied.item().clone())
    } else {
        None
    };

    let mut result = crate::data::config::read(name_span, &configuration)?;

    if let Some(v) = get {
        let key = v.to_string();
        let value = result
            .get(&key)
            .ok_or_else(|| ShellError::string(&format!("Missing key {} in config", key)))?;

        let mut results = VecDeque::new();

        match value {
            Tagged {
                item: Value::Table(list),
                ..
            } => {
                for l in list {
                    results.push_back(ReturnSuccess::value(l.clone()));
                }
            }
            x => results.push_back(ReturnSuccess::value(x.clone())),
        }

        return Ok(results.to_output_stream());
    }

    if let Some((key, value)) = set {
        result.insert(key.to_string(), value.clone());

        config::write(&result, &configuration)?;

        return Ok(stream![Tagged::from_simple_spanned_item(
            Value::Row(result.into()),
            value.span()
        )]
        .from_input_stream());
    }

    if let Tagged {
        item: true,
        tag: Tag { span, .. },
    } = clear
    {
        result.clear();

        config::write(&result, &configuration)?;

        return Ok(stream![Tagged::from_simple_spanned_item(
            Value::Row(result.into()),
            span
        )]
        .from_input_stream());
    }

    if let Tagged {
        item: true,
        tag: Tag { span, .. },
    } = path
    {
        let path = config::default_path_for(&configuration)?;

        return Ok(stream![Tagged::from_simple_spanned_item(
            Value::Primitive(Primitive::Path(path)),
            span
        )]
        .from_input_stream());
    }

    if let Some(v) = remove {
        let key = v.to_string();

        if result.contains_key(&key) {
            result.swap_remove(&key);
            config::write(&result, &configuration)?;
        } else {
            return Err(ShellError::string(&format!(
                "{} does not exist in config",
                key
            )));
        }

        let obj = VecDeque::from_iter(vec![Value::Row(result.into()).simple_spanned(v.span())]);
        return Ok(obj.from_input_stream());
    }

    return Ok(vec![Value::Row(result.into()).simple_spanned(name)].into());
}
